import os
import random

class City:

  def __init__(self, name, neighbors):
    self.name = name
    self.neighbors = neighbors

class Board:

  def __init__(self):
    self.cities = []

    # NorthEast (8 cities)
    neighbors = "Quebec", "Montreal", "Boston"
    self.cities.append(City("Bangor", neighbors))
    neighbors = "Bangor", "Montreal"
    self.cities.append(City("Quebec", neighbors))
    neighbors = "Quebec", "Bangor", "Boston", "Albany", "Buffalo"
    self.cities.append(City("Montreal", neighbors))
    neighbors = "Bangor", "Montreal", "Albany", "New York"
    self.cities.append(City("Boston", neighbors))
    neighbors = "Buffalo", "Montreal", "Boston", "New York"
    self.cities.append(City("Albany", neighbors))
    neighbors = "Detroit", "Montreal", "Albany", "New York", "Philadelphia", "Pittsburgh"
    self.cities.append(City("Buffalo", neighbors))
    neighbors = "Buffalo", "Albany", "Boston", "Philadelphia"
    self.cities.append(City("New York", neighbors))
    neighbors = "Buffalo", "New York", "Norfolk", "Baltimore", "Pittsburgh"
    self.cities.append(City("Philadelphia", neighbors))

    # Appalachia (8 cities)
    neighbors = "Buffalo", "Pittsburgh", "Cincinnati", "Chicago"
    self.cities.append(City("Detroit", neighbors))
    neighbors = "Buffalo", "Philadelphia", "Baltimore", "Lewisburg", "Cincinnati", "Detroit"
    self.cities.append(City("Pittsburgh", neighbors))
    neighbors = "Philadelphia", "Norfolk", "Lewisburg", "Pittsburgh"
    self.cities.append(City("Baltimore", neighbors))
    neighbors = "Norfolk", "Charleston", "Atlanta", "Knoxville", "Lewisburg"
    self.cities.append(City("Charlotte", neighbors))
    neighbors = "Lewisburg", "Charlotte", "Atlanta", "Nashville", "Cincinnati"
    self.cities.append(City("Knoxville", neighbors))
    #neighbors = "Detroit", "Pittsburgh", "Lewisburg", "Knoxville", "Nashville", "St Louis"
    neighbors = "Detroit", "Pittsburgh", "Lewisburg", "Knoxville", "Nashville"
    self.cities.append(City("Cincinnati", neighbors))
    #neighbors = "Detroit", "Cincinnati", "St Louis", "Davenport", "St Paul", "Duluth"
    neighbors = "Detroit", "Cincinnati"
    self.cities.append(City("Chicago", neighbors))
    neighbors = "Pittsburgh", "Baltimore", "Norfolk", "Charlotte", "Knoxville", "Cincinnati"
    self.cities.append(City("Lewisburg", neighbors))

    # Atlantic Coast (3 cities)
    neighbors = "Philadelphia", "Charleston", "Charlotte", "Lewisburg", "Baltimore"
    self.cities.append(City("Norfolk", neighbors))
    neighbors = "Norfolk", "Savannah", "Atlanta", "Charlotte"
    self.cities.append(City("Charleston", neighbors))
    neighbors = "Charleston", "Jacksonville", "Mobile", "Atlanta"
    self.cities.append(City("Savannah", neighbors))

    # The South (4 cities)
    #neighbors = "Cincinnati", "Knoxville", "Atlanta", "New Orleans", "Little Rock", "St Louis"
    neighbors = "Cincinnati", "Knoxville", "Atlanta", "New Orleans"
    self.cities.append(City("Nashville", neighbors))
    neighbors = "Charlotte", "Charleston", "Savannah", "Mobile", "Nashville", "Knoxville"
    self.cities.append(City("Atlanta", neighbors))
    neighbors = "Atlanta", "Savannah", "Jacksonville", "Tampa", "New Orleans"
    self.cities.append(City("Mobile", neighbors))
    #neighbors = "Nashville", "Mobile", "Houston", "Little Rock"
    neighbors = "Nashville", "Mobile"
    self.cities.append(City("New Orleans", neighbors))

    # Florida (3 cities)
    neighbors = "Savannah", "Miami", "Tampa", "Mobile"
    self.cities.append(City("Jacksonville", neighbors))
    neighbors = "Jacksonville", "Tampa"
    self.cities.append(City("Miami", neighbors))
    neighbors = "Jacksonville", "Miami", "Mobile"
    self.cities.append(City("Tampa", neighbors))

  def getCityByName(self, name):
    for city in self.cities:
      if city.name == name:
        return city
    print("ERROR: city '" + name + "' is not defined")
    return 0

  # Returns the length of the shortest way between two cities
  def distance(self, start, target):
    if start == target:
      return 0

    dist = 1
    cities = [start]
    passedCities = []

    #print("start = " + start.name + " ; target = " + target.name)
    while True:
      #print("dist = " + str(dist))
      for city in cities:
        #print("city = " + city.name)
        if target.name in city.neighbors:
          return dist
      passedCities.extend(cities)
      newCities = []
      for city in cities:
        for neighbor in city.neighbors:
          if neighbor not in passedCities:
            newCities.append(self.getCityByName(neighbor))
      cities = newCities
      dist = dist + 1

class MamaOConnel:

  def __init__(self):
    self.board = Board() # TODO Board should be a "static" or "constant" class, see what we can do
    self.location = self.board.cities[random.randint(0, len(self.board.cities) - 1)]
    # We need at least 3 accesses to the city, to provide hints through different axes (but even cities with 3 accesses
    # can generate a triangle (like Bangor or Tampa), hence the need for 4 accesses)
    # It could be more proper to have a dedicated method to check if a city can be the location (using distance?)
    # Moreover, due to their location and their corner neighbors, Montreal and Jacksonville can never have enough hints
    # TODO Buffalo sometimes runs in an infinte loop as well; check if still the case after we finished adding cities
    while len(self.location.neighbors) < 4 or self.location.name in ("Montreal", "Jacksonville"):
      self.location = self.board.cities[random.randint(0, len(self.board.cities) - 1)]

    hints = [], [], []
    hintsDistance = 3, 3, 2 # TODO randomize
    for h in range(0, 3):


      # Next hop(s) of hint
      #for i in range(0, hintsDistance[h] - 1):
      i = -1
      while i < hintsDistance[h] - 1:
        if i == -1:
          # First hop of hint
          city = self.board.getCityByName(self.location.neighbors[random.randint(0, len(self.location.neighbors) - 1)])
          if h == 1:
            while city in hints[0]:
              city = self.board.getCityByName(self.location.neighbors[random.randint(0, len(self.location.neighbors) - 1)])
          if h == 2:
            while city in hints[0] or city in hints[1]:
              city = self.board.getCityByName(self.location.neighbors[random.randint(0, len(self.location.neighbors) - 1)])
          hints[h].insert(0, city)
          i = 0
        else:
          do = True
          city = self.board.getCityByName(hints[h][i].neighbors[random.randint(0, len(hints[h][i].neighbors) - 1)])
          counter = 0
          while self.board.distance(city, self.location) < i + 2 or h >= 1 and city in hints[0] or h == 2 and city in hints[1]:
            if counter > len(hints[h][i].neighbors) + 5: # arbitrary failsafe
              i = i - 1
              do = False
              hints[h].pop()
              break
            city = self.board.getCityByName(hints[h][i].neighbors[random.randint(0, len(hints[h][i].neighbors) - 1)])
            counter = counter + 1
          if do:
            hints[h].insert(i + 1, city)
            i = i + 1

  # at the end of generation, board and cities should be dropped (no use to keep them in saved data)

myMama = MamaOConnel()
